/*
       The Travelling Salesman Problem Solver

       City Database
       Runs on both host and Hypercube.
*/	
#include <assert.h>
#include "config.h"
#include "travel.h"
extern long lsqrt();		       /* external */
extern boolean taxicab;		       /* taxicab metric */
/*
       Static Variables
*/
static boolean valid = FALSE;	       /* not valid until loaded */
static struct cities citydb;	       /* table	of city	information */
static long citydist[NCITIES][NCITIES];/* table	of inter-city distances	*/
/*
       setcities  --  set data into city database

       This copies a city database constructed elsewhere into one
       local structure.	 
*/
setcities(cityp)
struct cities *cityp;		       /* city database	IN */
{      short cnt = cityp->cccount;     /* count	of cities */
       short i,j;		       /* for loops */
       if (cnt > NCITIES) panic("setcities: too big");
       for (i=0; i<cnt;	i++)
       {       citydb.ccc[i].cityx = cityp->ccc[i].cityx; /* X coord */
	       citydb.ccc[i].cityy = cityp->ccc[i].cityy; /* Y coord */
       }
       /*      Clear table of precalcuated distances.  */
       for (i=0; i<NCITIES; i++)
	       for (j=0; j<NCITIES; j++)
		       citydist[i][j] =	-1; /* clear */
       citydb.cccount =	cnt;	       /* set count */
       valid = TRUE;		       /* now valid */
} 
/*
       getcity	--  get	information about one city

       An access procedure.  Not used extensively.
*/
getcity(cindex,x,y)
short cindex;			       /* city index IN	*/
long *x, *y;			       /* coordinates OUT */
{      assert(valid);
       assert(cindex >=	0 && cindex < citydb.cccount);
       *x = citydb.ccc[cindex].cityx;  /* return X */
       *y = citydb.ccc[cindex].cityy;  /* return Y */
}
/*
       getcnt  --  get count of	cities

       An access procedure.  Not used extensively.
*/
int getcnt()
{      assert(valid);		       /* database must	be loaded */
       return(citydb.cccount);	       /* return count */
}
/*
       distance	 --  distance between two cities

       Assumes a rectangular coordinate	system.
       If "taxicab" is set, the distance is simply dx + dy, the
       so called "taxicab metric", for grid-pattern cities!
*/
long distance(c1,c2)		       /* returns distance */
short c1,c2;			       /* city indices OUT */
{      long dx,dy;		       /* change in X and Y */
       long dist;		       /* distance */
       long distsq;		       /* distance squared */
       assert(valid);		       /* database must	be loaded */
       assert(c1 != c2);	       /* must be different cities */
       assert(c1>=0 && c1<citydb.cccount); /* valid city number	*/
       assert(c2>=0 && c2<citydb.cccount); /* valid city number	*/
       dist = citydist[c1][c2];	  /* look up if	precalculated */
       if (dist	>= 0) return(dist);    /* already did it, good */
       dx = (long)citydb.ccc[c1].cityx - citydb.ccc[c2].cityx;
       dy = (long)citydb.ccc[c1].cityy - citydb.ccc[c2].cityy;
       if (taxicab)		       /* if taxicab distance metric */
       {       dist =(abs(dx) +	abs(dy));  /* cheap calc */
       } else {			       /* if true distance metric */
	       distsq =	dx*dx +	dy*dy; /* distance squared */
	       dist = lsqrt(distsq);  /* calculate distance */
       }
       citydist[c1][c2]	= dist;	  /* save for reuse */
       return(dist);		       /* return distance */
}
