/*
 * $Header: cluster.c,v 1.7 90/11/12 20:31:22 stolcke Exp Locker: stolcke $
 * $Log:	cluster.c,v $
 * Revision 1.7  90/11/12  20:31:22  stolcke
 * two initialization fixed
 * 
 * Revision 1.6  90/11/12  18:43:26  stolcke
 * workaround to avoid old Sun compiler bug
 * 
 * Revision 1.5  90/11/12  16:40:34  stolcke
 * added metric selection and scaling
 * 
 * Revision 1.4  90/11/11  18:14:27  stolcke
 * -n flag no longer needed
 * 
 * Revision 1.3  90/11/11  02:30:14  stolcke
 * no more seek needed, can use in pipe now
 * 
 * Revision 1.2  90/11/10  23:44:02  stolcke
 * added new features (see manpage)
 * 
 */
#include <stdio.h>
#include <math.h>
#include "alloc.h"
#include "error.h"

typedef struct _tree {
  float	*pat;
  int	size;
  int	root;
  int	leaf;
  float	y;
  float distance;
  struct _tree *r_tree, *l_tree;
} BiTree;

#define X0 0.01
#define XMAX 80
#define NONE (-2)
#define TRUE 1
#define FALSE -1
#define IfEOF( x ) if( (x)== EOF )
#define Lname 32
#define BUFSIZE 256

#define SCALE "_SCALE_"

float distance();
float root();
BiTree *new_tree();

static char buffer[BUFSIZE];	/* temporary string read buffer */

static int vflag = 0;	/* print explanatory messages */
static int dflag = 0;	/* print distances */
static int tflag = 0;	/* print tree in ASCII */
static int gflag = 0;	/* print tree in graph(1) format */
static int bflag = 0;	/* print tree in graph(1) format, with breaks */
static int sflag = 0;	/* suppress scaling  */

static int norm = 2;	/* which l-norm to use for distance metric */

usage( program )
     char *program;
{
     fprintf(stderr, "usage: %s [-dtgbv] [-n norm] [vectorfile [namesfile]]\n", program );
     exit(2);
}

main ( argc, argv )
     int argc; char *argv[];
{
  FILE *fp, *fopen();
  float **pattern = NULL;
  int	lpat, npat;
  char  **name = NULL;

  int opt;
  extern char *optarg;
  extern int optind;

  while ((opt = getopt(argc, argv, "dtgbvn:s")) != -1)
     switch (opt) {
	case 'd':
		dflag = 1;
		break;
	case 't':
		tflag = 1;
		break;
	case 'g':
		gflag = 1;
		break;
	case 'b':
		bflag = 1;
		gflag = 1;
		break;
	case 'v':
		vflag = 1;
		break;
	case 'n':
		if ( sscanf(optarg, "%d", &norm) != 1 || norm < 0 )
			usage(argv[0]);
		break;
	case 's':
		sflag = 1;
		break;
	case '?':
		usage(argv[0]);
		break;
     }

  if ( !(dflag || tflag || gflag) )
	dflag = tflag = vflag = 1;		/* default behavior */
	
  if ( optind+2 < argc )
	usage(argv[0]);

  if ( !(optind < argc) || !strcmp(argv[optind], "-") )
	fp = stdin;
  else IfErr ( fp = fopen ( argv[optind], "r" ) ) {
	fprintf ( stderr, "%s: cannot open file %s\n", argv[0], argv[optind] );
	exit(1);
  }

  IfErr( read_pattern ( fp,  &pattern, &lpat, &npat, &name ) ) {
	fprintf ( stderr, "%s: %s: cannot read pattern\n", argv[0], ERR_MSG);
	exit(1);
  }
  if ( vflag )
	printf ( "read %d patterns:  size = %d\n", npat, lpat );

  if ( optind + 1 < argc ) {
	name = new_array_of( npat, char* );

	if ( !strcmp(argv[optind+1], "-") )
		fp = stdin;
	else IfErr( fp = fopen( argv[optind+1], "r" ) ) {
		fprintf ( stderr, "%s: cannot open file %s\n", argv[0], argv[optind+1] );
		exit(1);
	}

	IfErr( read_names( fp, name, npat ) ) {
		fprintf( stderr, "%s: %s: cannot read names\n", argv[0], ERR_MSG);
		exit(1);
	}
  }
  cluster( pattern, name, lpat, npat );
}

read_names ( fp, name, npat )
     FILE *fp; char **name; int npat;
{
  register int i;

  for( i=0; i< npat ; i++ ) {
    IfEOF( fscanf ( fp, "%s", buffer ) ) Erreturn("not enough names");
    IfErr( name[i] = new_string( buffer )) Erreturn("not enough core");
  }
  return( OK );
}

names( tree, name, str )
     BiTree *tree; char **name, *str;
{
  if( tree->leaf != FALSE )  {
    if(name) sprintf( str, "%s %s", str, name[ tree->leaf ] );
    else sprintf( str, "%s %d", str, tree->leaf );
    return( 1 );
  }
  else return( names( tree->r_tree, name, str )
	      + names( tree->l_tree, name, str ));
}

cluster ( pattern, name, lpat, npat )
     float **pattern; char **name; int lpat, npat;
{
  register int i,j,k;
  float	   dist,min_dist;
  int	   pair1, pair2, n_names;
  BiTree   *item = new_array_of ( npat, BiTree );
  float    **dist_matrix = new_2d_array_of( npat, npat, float );
  char	   *names1 = new_array_of( npat*Lname, char );
  char	   *names2 = new_array_of( npat*Lname, char );
  for ( i=0; i< npat ; i++ ) {
    item[i].pat = new_array_of ( lpat, float );
    for ( j=0; j< lpat ; j++ ) item[i].pat[j] = pattern[i][j];
    item[i].root = TRUE;
    item[i].size = 1;
    item[i].leaf = i;
    item[i].distance = 0.0;
  }
  for( i=0; i < npat; i++ )
    for( j=0; j < i; j++ )
      dist_matrix[i][j] = distance( item[i].pat,item[j].pat, lpat );

  for ( ; ; ) {
    min_dist = 999999999.9;
    pair1 = NONE;
    for ( i=0; i < npat ; i++ ) 
      for ( j=0; j < i ; j++ ) {
	if ( item[i].root == FALSE || item[j].root == FALSE ) continue;
	if( (dist = dist_matrix[i][j]) < min_dist ) {
	  min_dist = dist ;
	  pair1 = i;
	  pair2 = j;
	}
      }
    if ( pair1 == NONE ) break;		/* analysis finished */
    min_dist = root( min_dist );
    *names1 = *names2 = NULL;
    n_names = names( &item[pair1], name, names1 ) +
	      names( &item[pair2], name, names2 );

    if ( dflag )
	if( n_names > 10 ) printf( "minimum distance = %f\n(%s )\n(%s )\n", 
			      min_dist, names1, names2 );
	else printf( "minimum distance = %f\t(%s )\t(%s )\n", 
		min_dist, names1, names2 );

    item[ pair1 ].l_tree = new_tree ( &item[ pair1 ], lpat ); /* copy */
    item[ pair1 ].r_tree = &item[ pair2 ];
    item[ pair1 ].leaf = FALSE;	/* ith item cannot be a leaf it has subtrees */
    for ( k=0; k< lpat ; k++ )  /* pat of non-leaf is weighted average of    *
				 * pat's of its right & left subtrees */
      item[pair1].pat[k] = 
	(item[pair1].pat[k]*item[pair1].size
	 + item[pair2].pat[k]*item[pair2].size )
	  / (item[pair1].size + item[pair2].size );
    item[pair1].size = item[pair1].size + item[pair2].size ;
    item[pair1].distance = min_dist ;

    item[pair2].root = FALSE;	/* jth item is no longer a root.its a subtree*/
    for( i = 0; i < pair1 ; i++ ) { 	/* compute dist b/w new and other pat*/
      if( item[i].root == FALSE ) continue;
      dist_matrix[pair1][i] = distance(item[pair1].pat, item[i].pat, lpat );
    }
    for( i = pair1+1 ; i < npat; i++ ) {
      if( item[i].root == FALSE ) continue;
      dist_matrix[i][pair1] = distance(item[pair1].pat, item[i].pat, lpat );
    }
  }
  for ( i=0; i<npat ; i++ ) 
    if ( item[i].root == TRUE ) break;  /* there should be only one root */

  if ( tflag ) {
	if ( vflag )
		printf ( "Resulting Tree = \n" );
	print_tree ( &item[i], name, npat );
  }

  if ( gflag ) {
	if ( vflag )
		printf ( "Tree Graph= \n" );
	graph_tree ( &item[i], name, npat );
  }
}

float
distance ( pat1, pat2, lpat )
	float *pat1, *pat2;
	int lpat;
{
	register int i;
	float dist = 0.0, max = 0.0;

  for ( i=0; i< lpat ; i++ ) {
	float diff = pat1[i]-pat2[i];

	switch (norm) {
	    	float adiff;

	    case 2:
		dist += diff * diff;
		break;
	    case 1:
		dist += fabs(diff);
		break;
	    case 0:
		if ( (adiff = fabs(diff)) > dist )
			dist = adiff;
		break;
	    default:
		dist += pow(fabs(diff), (double)norm);
		break;
	}
  }
  return ( dist );
}

float
root( dist )
	float dist;
{
	switch ( norm ) {
	case 2:
		return( sqrt(dist) );
	case 1:
	case 0:
		return( dist );
	default:
		return( pow(dist, 1/(double)norm) );
	}
}

BiTree *
new_tree ( item, lpat )
     BiTree *item;
{
  register int i;
  BiTree *tree = new ( BiTree );

  tree->r_tree = item->r_tree;
  tree->l_tree = item->l_tree;
  tree->leaf = item->leaf;
  tree->root = item->root;
  tree->distance = item->distance;
  tree->pat = new_array_of ( lpat, float );
  for (i=0; i< lpat ; i++ ) tree->pat[i] = item->pat[i];
  return ( tree );
}

max_lname( name, npat )
  char **name; int npat;
{
  register int i;
  int max = 0, lname;
  if( name == NULL ) return( 3 );
  for( i=0; i < npat; i++ )
    if( (lname = strlen( name[i] )) > max ) max = lname;
  return( max );
}

static float Y;

graph_tree ( tree, name, npat )
     BiTree *tree; char **name; int npat;
{
  register int i,j;
  float x0;
  float total_distance(), y_tree();
  int lname = max_lname( name, npat );  /* maximum name length */

  x0 = X0 * total_distance(tree);
  Y = 0.5;
  (void) y_tree( tree, 0 );
  graph_tree_walk( tree, name, -x0, x0 );
}

graph_tree_walk( tree, name, x, distance )
	BiTree *tree;
	char **name;
	float x, distance;
{
  float y = tree->y, x_next = x + distance;
  if( tree->leaf != FALSE ) {
	printf( "%f %f\n", x, y );
	printf( "%f %f", x_next, y );
	if( name == NULL )
		printf( " \" %d\"\n", tree->leaf );
	else
		printf( " \" %s\"\n", name[tree->leaf] );
	if ( !bflag )
		printf( "%f %f\n", x, y );
	else
		printf( "\n" );
  }
  else {
	float y_l = tree->l_tree->y, y_r = tree->r_tree->y;

	printf( "%f %f\n", x, y);
	printf( "%f %f\n", x_next, y);
	graph_tree_walk( tree->l_tree, name, x_next, tree->distance );
	if ( !bflag )
		printf( "%f %f\n", x_next, y);
	printf( "%f %f\n", x_next, y);
	graph_tree_walk( tree->r_tree, name, x_next, tree->distance );
	if ( !bflag ) {
		printf( "%f %f\n", x_next, y);
		printf( "%f %f\n", x, y);
	}
  }
}

print_tree ( tree, name, npat )
     BiTree *tree; char **name; int npat;
{
  register int i,j; char **space;
  float total_distance(), y_tree();
  int lname = max_lname( name, npat );  /* maximum name length */
  float xscale = (XMAX - tree_depth(tree)*2 -lname-3) / total_distance(tree) ;
  Y = 0.0;
  (void) y_tree( tree, 1 );
  space = new_2d_array_of( tree->size, XMAX, char );
  for( i=0; i< tree->size; i++ )
    for( j=0; j < XMAX ; j++ ) space[i][j] = ' ';
  draw_tree_in_space( tree, name, space, 0, xscale, 0.0 );
  print_space( space, XMAX, tree->size );
  free_2d_array( space, tree->size );
}

print_space( space, depth, size )
     char **space; int depth, size;
{
  register int i;
  for( i=0; i<size; i++ ) {
    space[i][depth-1] = NULL;
    puts( space[i] );
  }
}

draw_tree_in_space( tree, name, space, x, xscale, distance )
BiTree *tree; char **name; char **space; int x; float xscale, distance;
{
  int y_l,y_r, y = (int) tree->y; char char_x; int x_next;
  if( tree->leaf != FALSE ) {
    for( x_next = x + distance * xscale; x <= x_next; x++ )
      space[y][x] = '-';
    if( name == NULL ) sprintf( &space[y][x], "> %d", tree->leaf );
    else sprintf( &space[y][x], "> %s", name[tree->leaf] );
  }
  else {
    y_l = (int) tree->l_tree->y; y_r = (int) tree->r_tree->y;
    if( (y_l - y_r) < 2 ) {
      if( x > 0 ) space[y][x-1] = ' ';
      char_x = '_';
    }
    else  char_x = '-';
    for( x_next = x + distance * xscale; x <= x_next; x++ ) 
      space[y][x] = char_x;

    draw_tree_in_space( tree->l_tree, name, space, x+1, xscale, tree->distance );
    for( y = y_r ; y <= y_l ; y++ ) space[y][x] = '|';
    draw_tree_in_space( tree->r_tree, name, space, x+1, xscale, tree->distance );
  }
}
#if 0
draw_tree_in_space( tree, name, space, x, xscale )
     BiTree *tree; char **name; char **space; int x; float xscale;
{
  int y_l,y_r, y = (int) tree->y; char char_x; int x_next;
  if( tree->leaf != FALSE ) {
    if( name == NULL ) sprintf ( &space[y][x], "---> %d", tree->leaf );
    else sprintf( &space[y][x], "---> %s", name[tree->leaf] );
  }
  else {
    y_l = (int) tree->l_tree->y; y_r = (int) tree->r_tree->y;
    if( (y_l - y_r) < 2 ) {
      if( x > 0 ) space[y][x-1] = ' ';
      char_x = '_';
    }
    else  char_x = '-';
    for( x_next = x + tree->distance * xscale; x <= x_next; x++ ) 
      space[y][x] = char_x;

    draw_tree_in_space ( tree->l_tree, name, space, x+1, xscale );
    for( y = y_r ; y <= y_l ; y++ ) space[y][x] = '|';
    draw_tree_in_space ( tree->r_tree, name, space, x+1, xscale );
  }
}
#endif 0
tree_depth( tree )
     BiTree *tree;
{
  int depth_r, depth_l;
  if( tree->leaf == FALSE ) 
    return( ((depth_r=tree_depth( tree->r_tree )) >
	     (depth_l=tree_depth( tree->l_tree )))? depth_r+1 : depth_l+1 );
  else return( 1 );
}

float
total_distance( tree )
     BiTree *tree;
{
  float dist_r, dist_l, total_distance();
  if( tree->leaf == FALSE ) 
    return( ((dist_r=total_distance( tree->r_tree )) >
	     (dist_l=total_distance( tree->l_tree )))? 
	    tree->distance+dist_r : tree->distance+dist_l );
  else return( tree->distance );
}

float
y_tree ( tree, right_to_left )
     BiTree *tree;
     int right_to_left;
{
  if( tree->leaf == FALSE ) {
        float y_r, y_l;

	if ( right_to_left ) {
		y_r = y_tree( tree->r_tree, right_to_left);
		y_l = y_tree( tree->l_tree, right_to_left);
	} else {
		y_l = y_tree( tree->l_tree, right_to_left);
		y_r = y_tree( tree->r_tree, right_to_left);
	}
	return( tree->y = (y_r + y_l)/2 );
  }
  else {
	float y = Y;

	Y += 1.0;
	return( tree->y = y );
  }
}

#define Blksize 128

read_pattern( Pfile, patternP, lpatternP, npatternP, nameP )
     FILE	*Pfile; 			/* ptrs to pattern file */
     float	***patternP;
     int	*lpatternP, *npatternP;
     char	***nameP;
{
  register int	i;
  int   status;
  int	Asize = Blksize ;		/* current array size */

        /***** these local variables are temporary storage and are ****
	 ***** copied to the real variables if there is no error ******/
  int	lpattern, npattern;		/* size and # of patterns */
  float	**pattern, x;
  float *first_pattern;
  char	**name = NULL;
  char	*first_name = NULL;
  float *scales = NULL;

  IfErr( lpattern = nstrings ( Pfile, &first_pattern, &first_name ) )
	return( ERR );

  /* check for scaling info */
  if ( first_name != NULL && strcmp(first_name, SCALE) == 0 ) {
	scales = first_pattern;
	free(first_name);
	
	/* read first vector again */
	IfErr( status = nstrings ( Pfile, &first_pattern, &first_name ) )
		return( ERR );
	if ( status != lpattern )
		Erreturn( "scaling vector not matching data" );
  }

  /* check for scaling info */

				/* allocate space for input/target arrays */
  IfErr( pattern = new_2d_array_of ( Asize, lpattern, float ) ) 
    Erreturn("cannot allocate memory for patterns" );
 
  if ( first_name != NULL )
	if Err( name = new_array_of( Asize, char* ) ) 
	  Erreturn("cannot allocate memory for names" );
 
	/**** this loop reads in one line from pattern file,**
	 **** stores each pattern into pattern buffer    *****/

  for( npattern=0;  ; npattern++ ) { 

    if( npattern >= Asize ) {     /* need to allocate more space for arrays */
      IfErr( pattern = change_2d_array_size(
	    pattern, Asize, lpattern, Asize+Blksize, lpattern, float ))
	Erreturn( "cannot allocate memory for pattern " ); 

      if ( name != NULL )
	IfErr( name = change_array_size( name, Asize+Blksize, char* ) )
	  Erreturn( "cannot allocate memory for pattern " ); 

      Asize += Blksize;	/* array size is now Blksize bigger */
    }
	
    if ( first_pattern != NULL ) {	/* copy data from first line */
	for ( i=0; i< lpattern ; i++ )
		pattern[npattern][i] = first_pattern[i];
	free(first_pattern);
	first_pattern = NULL;

	if ( first_name != NULL ) {
		name[npattern] = first_name;
		first_name = NULL;
	}
    } else {				/* read data from file */
	for ( i=0; i < lpattern ; i++ ) {
	    IfEOF ( status = fscanf (Pfile, "%f", &pattern[npattern][i] ) ) {
		if ( i==0 ) break;
		Erreturn1("cannot read pattern # %d", npattern ); 
	    }
	    IfErr ( status )
		Erreturn1("cannot read pattern # %d", npattern); 
	}
	IfEOF ( status ) break;

	if ( name != NULL ) {
		IfEOF( status = fscanf ( Pfile, "%s", buffer ) )
		  Erreturn("not enough names");
		IfErr( name[npattern] = new_string( buffer ))
		  Erreturn("not enough core");
	}
	IfEOF ( status ) break;
    }

    if ( !sflag && scales != NULL )		/* scale data if requested */
	for ( i=0; i< lpattern ; i++ )
		pattern[npattern][i] *= scales[i];

  }
	/* if there is any error, these pointers below aren't changed */

		/* free any space already allocated for patterns */
  if( *patternP != NULL ) free_2d_array ( *patternP, *npatternP );
  if( *nameP != NULL ) free ( *nameP );

  *npatternP = npattern;	/* # of patterns read in from file */
  *lpatternP = lpattern;	/* # of elements in pattern */
  *patternP = pattern;		/* input array */
  *nameP = name;		/* name array */

  return( OK );		/* patterns were read in without error */
}

nstrings ( fp, patternP, nameP ) /* counts number of strings in the first line of file*/
  FILE *fp;
  float **patternP;
  char **nameP;
{
  register int i;
  int	c;
  float	*pattern;
  char	*name = NULL;

  pattern = new_array_of(1, float );

  for( i=0; ; i++ ) {
	/* skip blanks and look for end of line */
	while( (c = getc ( fp ))==' ' || c == '\t' ); 
	if( c=='\n' ) break;
	ungetc( c, fp );

	/* read field (number of name) */
	if ( fscanf(fp, "%s", buffer) != 1 ) break;

	if ( sscanf(buffer, "%f", &pattern[i]) != 1) {
		name = new_string(buffer);
		while ( (c = getc(fp)) != '\n' && c != EOF );	/* skip to next line */
		break;
	}

	IfErr( pattern = change_array_size(pattern, (i+2), float) )
	  Erreturn("not enough core");
  }

  if ( i == 0 ) Erreturn("empty pattern");

  *patternP = pattern;
  *nameP = name;
  return( i );
}
