


LP_SOLVE(1ES)       UNIX Programmer's Manual        LP_SOLVE(1ES)



NAME
     lp_solve  - Solve (mixed integer) linear programming prob-
     lem.

SYNOPSIS
     lp_solve [option]* "<" <input-file>

OPTIONS
     -v          Verbose mode

     -h          Help mode, prints the usage

     -d          Debug mode, all intermediate results are
                 printed, and the branch-and-bound decisions in
                 case of (mixed) integer problems.

     -b <bound>  Specify a lower limit for the value of the
                 objective function to the program. Only useful
                 for (mixed) integer problems.  If close enough,
                 may speed up the calculations.

     -c          When branching in MILP problems, take the ceil-
                 ing of the selected non-integer variable first
                 instead of the floor. This can influence the
                 speed of MILP problems.

     -e <value>  Specify the accuracy with which it is checked
                 whether the value of a variable is really
                 integer. <value> must be between 0 and 0.5.
                 Default value is 1e-6 and should be ok for most
                 applications. Of course only useful for MILP
                 problems.

     -i          Print all intermediate valid solutions. Can give
                 you useful solutions even if the total run time
                 is too long.  Only useful for (mixed) integer
                 problems.

DESCRIPTION
     The linear programming problem can be formulated as: Solve
     A.x >= V1, with V2.x maximal. A is a matrix, x a vector of
     variables, V1 a vector called the right hand side, and V2 a
     vector specifying the objective function.
     Any of the variables may be specified integer.
     This program solves problems of this kind. It is slightly
     more general than the above problem, in that every row of A
     (specifying one equation) can have its own (in)equality, <=,
     >= or =. The result specifies values for all variables.
     Uses a 'Simplex' algorithm and sparse matrix methods, for
     pure LP problems.  If one or more of the variables is
     declared integer, the Simplex algorithm is iterated with a
     branch and bound algorithm, until the desired optimal



Printed 10/8/93                                                 1






LP_SOLVE(1ES)       UNIX Programmer's Manual        LP_SOLVE(1ES)



     solution is found.

INTERRUPT
     In order to make it possible to get useful solutions for a
     mixed integer problem which takes too long to finish com-
     pletely, the "interrupt" signal (kill -INT, ^C for many peo-
     ple) will not terminate the program when it is solving a
     MILP problem, but rather print the best valid solution found
     so far. This feature is not useful and thus disabled when
     you are solving a pure LP problem. If you see all zeros, it
     implies that no valid solution has been found yet. To really
     terminate the program, use a different signal, like KILL.

INPUT SYNTAX
     The input syntax is a set of algebraic expressions and "int"
     declarations in the following order:

     <objective function>
     <constraint>+
     <declaration>*

     where:

     - <objective function> is a linear combination of variables,
       ending with a semicolon, optionally preceded by "max: " or
       "min: " to indicate whether you want it to be minimized or
       maximized. The case is not important, "Max:" or "MAX:"
       will work as well. Maximization is the default.

     - <constraint> is a linear combination of variables and con-
       stants, followed by a relational operator, followed again
       by a linear combination of variables and constants, ending
       with a semicolon. The relational operator can be any of
       the following: "<" "<=" "=" ">" ">=". There is no semantic
       difference between "<" and "<=" nor between ">" and ">=".

     - <declaration> is of the form: "int" <var>+ ";" Commas are
       allowed between variables.

       So, the simplest linear problem consists of an objective
       function and 1 constraint.

EXAMPLE
     The simple problem:

     x1 >= 1
     x2 >= 1
     minimise x1 + x2 (= maximise -(x1 + x2)), with x1 integer

     can be written as follows:

     -x1 + -x2;



Printed 10/8/93                                                 2






LP_SOLVE(1ES)       UNIX Programmer's Manual        LP_SOLVE(1ES)



     (or min: x1 + x2;)
     x1 > 1;
     x2 > 1;
     int x1;

     The correct result for (x1, x2) is of course (1, 1).

BUGS
     There are a few situations where lp_solve is known to give
     wrong "No solutions".

     -    The problem consists entirely of constraints on just
          single variables (so-called "bounds", like x < 1; ) and
          no constraint with more than 1 variable (like x + 3 y >
          17; ). This leaves lp_solve with an empty problem
          matrix, as bounds are not stored in the main matrix. No
          real-life examples should be of this form, so I am not
          really chasing this problem.

     -    There are variables in the objective function which do
          not occur in a constraint with more than 1 variable,
          but just in a bound. This is a bug in lp_solve which I
          would like to repair, but I have sofar been unsuccess-
          ful.  Anyway, this can easily be avoided, as the value
          of a variable just occurring in a bound and the objec-
          tive function is trivially clear from the beginning.
          You can also just constrain the value of the entire
          objective function in the first constraint by a number
          which you know will not constrict your LP problem.

     -    Many people forget that lp_solve can only handle POSI-
          TIVE values for the variables.

     -    Sometimes problems are numerically unstable, and the
          unavoidable rounding errors inside lp_solve will cause
          wrong results. It is very hard to give general solu-
          tions to this problem, but try to keep all values in
          your problem in the order of magnitude of 1 by proper
          scaling. Almost parallel constraints are also not very
          good for numerical stability. Use "lp_solve -v" and
          observe the values of the pivots to see if there are
          any dangerously large or low numbers there.

SEE ALSO
     The implementation of the simplex kernel was mainly based
     on:
     W. Orchard-Hays: "Advanced Linear Programming Computing
     Techniques", McGraw-Hill 1968

CONTRIBUTED BY
     M.R.C.M. Berkelaar
     Eindhoven University of Technology



Printed 10/8/93                                                 3






LP_SOLVE(1ES)       UNIX Programmer's Manual        LP_SOLVE(1ES)



     Design Automation Section
     P.O. Box 513
     NL-5600 MB Eindhoven, The Netherlands
     phone ...-31-40-473345
     E-mail: michel@es.ele.tue.nl

STATUS
     Use at own risk. Bug reports are welcome, as well as succes
     stories.














































Printed 10/8/93                                                 4



