/*
       The Travelling Salesman Problem Solver

       Read Geography Information

       This module reads the geography database	for a file.

       This module executes on the host	machine	only.
*/	
#include "stdio.h"
#include "config.h"
#include "travel.h"
extern char *malloc();		       /* from library */
/*
       rgeo  --	 read a	geography database

       A geography database consists of	lines of the form

       <X coordinate> <Y coordinate> <name>
*/
int rgeo(fname,cityp,namep)	       /* returns 0 if success,	else -1	*/
char fname[];			       /* filename to read */
struct cities *cityp;		       /* city structure to fill (OUT) */
char *namep[];			       /* names	of cities (OUT)	*/
{      FILE    *fgeo;		       /* geography file */
       int stat;		       /* returned status */
       short badinput =	0;	       /* count	of bad input lines */
       short lineno = 0;	       /* line counter */
       fgeo = fopen(fname,"r");        /* open for reading */
       if (fgeo	== NULL)	       /* fails	*/
       {       fprintf(stderr,"Unable to open geography file %s\n",
		       fname);
	       return(-1);	       /* fails	*/
       }
       stat = rgeorecs(fgeo,cityp,namep); /* read the geometry records */
       fclose(fgeo);		       /* always close the file	*/
       return(stat);		       /* returns status */
}
/*
       rgeorecs	 --  read a geography database

       A geography database consists of	lines of the form

       <X coordinate> <Y coordinate> <name>
*/
int rgeorecs(fgeo,cityp,namep)	       /* returns 0 if success,	else -1	*/
FILE *fgeo;			       /* file to read */
struct cities *cityp;		       /* path structure to fill (OUT) */
char *namep[];			       /* names	of cities (OUT)	*/
{     
       short badinput;		       /* count	of bad input lines */
       short lineno;		       /* line counter */
       long x,y;		       /* coords */
       char cityname[120];	       /* name of city */
       char line[120];		       /* input	line */
       badinput	= 0; lineno = 0;       /* clear	counters */
       cityp->cccount =	0;	       /* no cities yet	*/
       for (;;)			       /* main read loop */
       {       if (fgets(line,sizeof(line),fgeo) == NULL) break;
	       lineno++;	       /* tally	lines */
	       x = -1;		       /* initialize to	invalid	values */
	       y = -1;
               cityname[0] = '\0';         
               sscanf(line,"%ld %ld %s",&x,&y,cityname);
               if (x < 0 || y < 0 || cityname[0] == '\0' 
	       || x > 32767 || y > 32767)
               {       fprintf(stderr," Invalid line in geography file:\n  %d.   %s\n",
			       lineno,line);
		       badinput++;     /* tally	bad lines */
		       if (badinput > 10) return(-1); /* fails if too many */
		       continue;       /* try reading next line	*/
	       }
	       if (addcity((short)x,(short)y,cityname,cityp,namep)) /* add city	*/
	       {       badinput++;     /* error	*/
		       break;
	       }
       }
       if (badinput) return(-1);       /* fails	if any bad input */
       return(0);		       /* success */
}
/*
       addcity	--  add	a city to the database
*/
static int addcity(x,y,cityname,cityp,namep) /*	returns	-1 if fail */
short x,y;			       /* loc of city */
char cityname[];		       /* name of city */
struct cities *cityp;		       /* city structure */
char *namep[];			       /* name of city */
{
       short i;			       /* for loop */
       short len;		       /* length of city name, less null */
       char *np;		       /* city name buffer */
       if (cityp->cccount >= NCITIES)  /* if table full	*/
       {       fprintf(stderr,"Too many cities, limit %d\n",NCITIES);
	       return(-1);	       /* fails	*/
       }
       len = strlen(cityname);	       /* length of city name */
       np = (char *)malloc(len+1);     /* allocate space for name */
       for (i=0; i<=len; i++)	       /* copy name of city */
       {       np[i] = cityname[i]; }  /* copy including null */
       namep[cityp->cccount] = np;     /* save name in name table */
       cityp->ccc[cityp->cccount].cityx	= x; /*	X location */
       cityp->ccc[cityp->cccount].cityy	= y; /*	Y location */
       cityp->cccount++;	       /* advance to next entry	*/
       return(0);		       /* done */
}
/*
       printcities  --	print the city database
*/
int printcities(namep)		 
char *namep[];			       /* name of city */
{      short i;			       /* for loop */
       short cnt;		       /* number of cities */
       long x,y;		       /* location of city */
       cnt = getcnt();		       /* get number of	cities */
       for (i=0; i<cnt;	i++)	       /* for all cities */
       {       getcity(i,&x,&y);       /* get city information */
               printf("%4d. (%ld,%ld) %s\n",i,x,y,namep[i]);
       }
}
