/*
       The Travelling Salesman Problem Solver

       Main Program
       Runs on host, not Hypercube.
*/	
#include "stdio.h"
#include "config.h"
#include "travel.h"
extern long distance();		       /* external */
extern char *utedit();		       /* time edit */
extern long utend();		       /* elapsed time */
/*
       Static Variables
*/
static char *nametab[NCITIES];	       /* pointers to names of cities */
static struct path pathtab;	       /* table	of path	information */
       boolean verbose = FALSE;	       /* verbosity flag */
       boolean debug = FALSE;	       /* debug	flag */
       boolean taxicab = FALSE;	       /* taxicab algorithm */
#define	CSTKSIZE  8000		       /* ***TEMP*** */
int _stack = CSTKSIZE;		       /* MS-DOS stack size */
/*
       getcityinfo  --	get city information

       The locations go	into the city database but the names are
       just kept in a local table.  Later, only	the locations will
       be propagated through the Hypercube.
*/
static void getcityinfo(fname)
char fname[];			       /* name to get from */
{      struct cities citytab;	       /* scratch city table */
       if (rgeo(fname,&citytab,nametab)) exit(1);
       setcities(&citytab);	       /* set in permanent database */
}
/*
       printpath  --  print a path structure, with city	names.
*/
void printpath(fp,pathp)
FILE *fp;				       /* write	to this	*/
struct path *pathp;			       /* path structure IN */
{
       short pos;			       /* current city */
       short i;				       /* for loop */
       short prev = NONE;		       /* no previous city yet */
       for (i=0; i<pathp->pathn; i++)	       /* for all on path */
       {       pos = pathp->plinks[i];	       /* current city */
               printf("%4d.  %s",pos,nametab[pos]);
	       if (prev	!= NONE)
               {       printf("  [%ld]",distance(prev,pos)); }
               printf("\n");
	       prev = pos;		       /* advance */
       }
       printf("Total length: %ld\n",pathp->pathl);
}
/*
       Main Program
*/
void main(argc,argv)
int argc;
char *argv[];			       /* args */
{      char *filename =	NULL;	       /* filename arg */
       short i;			       /* for loop */
       long runtime;		       /* total	run time */
       for (i=1; i<argc; i++)	       /* for all args */
       {       if (argv[i][0] != '-') 
	       {       if (filename)   /* if already have filename */
		       {       filename	= NULL;/* force	error */
			       break;
		       }
		       filename	= argv[i]; /* get filename */
		       continue;       /* on to	next argument. */
	       }
	       switch (argv[i][1]) {   /* fan out on flag */
               case 'v':       verbose = TRUE; /* verbose on */
			       break;
               case 'd':       debug = TRUE;   /* debug on */
			       break;
               case 't':       taxicab = TRUE; /* taxicab metric on */
			       break;
	       default:
                       fprintf(stderr,"Unknown flag: %s\n",
			       argv[i]);
		       exit(1);
	       }
       }
       if (filename == NULL)
       {       fprintf(stderr,"Usage: travel <flags> <geography file>\n");
	       exit(1);		       /* fails	*/
       }

       getcityinfo(filename);	       /* get city information */
       printcities(nametab);	       /* print	city table */
       printf("Initializing path table.\n");
       initpath(&pathtab);	       /* do it	*/
       if (verbose)
       {       printpath(stdout,&pathtab);     /* print	it */
       }
       checkpath(&pathtab);	       /* check	it */
       utstart();		       /* start	timing */
       printf("Attempting improvement.\n");
       improve(&pathtab);	       /* try to improve */
       runtime = utend();	       /* end timing */
       printf("\n FINAL PATH\n\n");
       printpath(stdout,&pathtab);     /* final	path */
       printf("Done.  %d cities in %s\n",
	       pathtab.pathn,utedit(runtime));	     /*	print time */
       exit(0);			       /* normal exit */
}
