/*
       The Travelling Salesman Problem Solver

       General utilities
       Can run on host or Hypercube.
*/	
#include "config.h"
#include "travel.h"
extern long distance();		       /* external */
/*
       checkpath  --  check a path structure for internal consistency

       Fatal if	an error is found
       This uses a lot of stack	space.
*/
checkpath(pathp)
struct path *pathp;			       /* path structure IN */
{
       short i;				       /* for loops */
       short cnt;			       /* number of cities */
       long length;			       /* length of path so far	*/
       short prev;			       /* previous length */
       short pos;			       /* current city */
       boolean chk[NCITIES];		       /* table	of cities seen */
       cnt = getcnt();			       /* get count of cities */
       prev = NONE;			       /* no previous entry */
       length =	0;			       /* length so far	*/
       if (cnt > NCITIES) panic("checkpath: oversize count");
       if (cnt < 2) panic("checkpath: undersize count");
       for (i=0; i<cnt;	i++) chk[i] = FALSE;   /* initialize check table */
       for (i=0; i<cnt;	i++)		       /* check	for dups */
       {
	       pos = pathp->plinks[i];	       /* get current city */
               if (pos < 0) panic("checkpath: link < 0");
               if (pos >= cnt) panic("checkpath: link >= cnt");
               if (chk[pos]) panic("checkpath: duplicate");
	       chk[pos]	= TRUE;		       /* make true */
	       if (prev	!= NONE) length	+= distance(prev,pos); /* tally	dist */
	       prev = pos;		       /* advance */
       }
       /*      We don't have to check that all cities were represented,  */
       /*      because our check for duplicates	and bounds did that.	 */
       /*      Finally check the length.       */
       if (length != pathp->pathl) panic("checkpath: length");
} 
/*
       ssqrt.c	--  short integer square root

       Binary search algorithm;	there are better approaches.
*/
static short ssqrt(n)		       /* returns square root */
short n;			       /* value	to take	square root of */
{      short hi	= 181;		       /* high guess */
       short lo	= 0;		       /* low guess */
       short mid;		       /* middle value */
       if (n <=	0) return(0);	       /* handle exception */
       for (;;)			       /* binary search	*/
       {       mid = (lo + (hi - lo)/2); /* middle value, avoids overflow */
	       if (mid * mid  >	n) /* if too big */
	       {       hi = mid	- 1;   /* new upper limit */
	       } else {lo = mid;}      /* new lower limit */
	       if (lo >= hi-1)	       /* interval down	to two */
	       {       if (hi *	hi <= n) return(hi);  /* final chk */
		       return(lo);     /* low value is sqrt */
	       }
       }
}
/*
       lsqrt.c	--  long integer square	root

       Binary search algorithm;	there are better approaches.
*/
long lsqrt(n)			       /* returns square root */
long n;				       /* value	to take	square root of */
{      long hi = 65535;		       /* high guess */
       long lo = 0;		       /* low guess */
       long mid;		       /* middle value */
       if (n <=	0) return(0);	       /* handle exception */
       if (n < 181*181)		       /* if short value */
       {       return((long)ssqrt((short) n)); /* use short routine */
       }
       for (;;)			       /* binary search	*/
       {       mid = (lo + (hi - lo)/2); /* middle value, avoids overflow */
	       if (mid * mid  >	n) /* if too big */
	       {       hi = mid	- 1;   /* new upper limit */
	       } else {lo = mid;}      /* new lower limit */
	       if (lo >= hi-1)	       /* interval down	to two */
	       {       if (hi *	hi <= n) return(hi);  /* final chk */
		       return(lo);     /* low value is sqrt */
	       }
       }
}
