#include "xor.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h> 

extern double sqrt(), log(), log10(), pow();


/***========== These 2 are the original functions that works on
                      a range of [-0.5,+0.5]        

float Sigmoid(float x) 
{
  return ( (1.0 / (1.0 + AM_EXP((double)-x))) - 0.5);
}

float SigmoidPrime(float x)
{
  x = Sigmoid(x);
  return ( 0.25 - (x * x));
}
=======================================***/


/*---------------------------------------------------*/
/* The function considered here is tanh()            */
/*---------------------------------------------------*/

float Sigmoid(float x)               /*  [-1.0,1.0]  */
{
  return(tanh((double) x)); 

/***--or-- return ( (2.0 / (1.0 + AM_EXP((double)-x))) - 1.0); --***/
}


float SigmoidPrime(float x)
{
  x = Sigmoid(x);
  return(1.0 - (x * x));
}

/*-----------------------------------------------------------*/
/* The following is a piecewise-rational function (PWR):     */
/* PWR approximates the highly smooth and non-linear func    */
/* (tanh) which ranges over (-1,1). Also this PWR takes      */
/* a linear form in the [-x_0,x_0] intreval.                 */
/* f(x) = 1 - [(1-x_0) x_0^k]/ x^k                           */
/* alpha = k = 1,2,...                                       */
/* x_0 = k/(1+k)                                             */
/*---------------------------------------------------------------*/
/*-------------------------------------------------------------------*/
#define alpha 1.0                           /*  alpha = k            */
#define x_0  (1.0/2.0)    /* 0.5/(1+k) or 0.6/(1+k)or ... or k/(1+k) */
#define top ((1.0 - x_0) * x_0)             /*  =(1 - x_0) x_0^k     */
/*-------------------------------------------------------------------*/


float absol(float value)
{
  if (value < 0.0)
    return(-1.0 * value);
  else
    return(value);
}

/*--------------------------------------*/
float Pwrfunc(float x)
{
  float func;
  if (x == 0)
    return 0.0;
  if (x <= x_0 && x >= -x_0)      
    return (x);
  if (absol(x) > x_0)
    {
      func = 1.0 - (top / absol(x));
      if (x > x_0)               
	return (func);
      else
	{
	  if (x < -x_0)       
	    return (-func);
	}
    }
}


/*----------------------------------------------------------- */
/* f'(x) = k((1 - x_0)x_0^k/ x^(k+1))                         */
/* where k=1,2,... and x_0 = k/(1+k)                          */
/*----------------------------------------------------------- */

float PwrPrimefunc(float x)
{
  if (x == 0.0)
    return 0.0;
  if (x <= x_0 && x >= -x_0)      
    return (1.0);
  return(alpha * (top /absol(x * x)));    /* for k = 1 */
}

/*=======================================*/
/* Another piecewise-rational function takes on the following
   form  f(x) = x / (1 + |x|) .
-----------------------------------------*/

float PWR_2(float x)
{
  return(x / (1.0 + absol(x)));
}

/*---------------------------------------*/
/* The above function is differentiable everywhere with
   a derivative:
-----------------------------------------*/

float PWR_2_prime(float x)
{
  return(1.0 / ((1.0 + absol(x)) * (1.0 + absol(x))));
}

/*=======================================*/
float signum(float x)
{
  if (x > 0.0)
    return 1.0;
  else
    {
      if (x < 0.0)
	return -1.0;
      else
	{
	  if (x == 0)
	    return 0.0;
	}
    }
} 

float LinearAND(float x)
{
 return (log(absol(x) + 1.0) * signum(x));
}

float LinearANDprime(float x)
{
  return ((1.0 /(absol(x) + 1.0)) * signum(x));
}
/*================================================*/
/*================================================*/

float L2Error (float *target, float *output, float *credit, int n)
{
  float mag_squared = 0.0;
  int i;

  for(i=0;i<n;i++) {
    credit[i] = target[i] - output[i];
    mag_squared += credit[i] * credit[i];
  }/* end for */

  return( mag_squared / 2.0 );
}


float L1Error (float *target, float *output, float *credit, int n)
{
  float mag = 0.0;
  int i;

  for(i=0;i<n;i++) {
    float val, diff;

    diff = target[i] - output[i] ;

    if (diff == 0.0)
      val = 0.0 ;
    else
      val = ( diff < 0.0 ) ? -1.0 : 1.0 ;

    credit[i] = val;
    mag += AM_FABS(diff);
  }/* end for */

  return( mag );
}


/*
**----------------------------------------------------------------------
**
**  Asymmetric thresholded squared error
**
**----------------------------------------------------------------------
*/

#define DELTA 0.150

#define GI 1.384   /* Credit multiplier for an Inhibited node */
#define GA 1.384   /* Credit multiplier for an Active node */

#define GIdb2 0.692
#define GAdb2 0.692

float asymmError(float *target, float *output, float *credit, int nx)
{
  float e;
  double t, total_error = 0.0;
  unsigned int k;
  
  for (k=0; k<nx; ++k) {
    
    e = output[k] - target[k];    
    
    if ( e < -DELTA ) {
      t = e + DELTA;
      credit[k] = -GA * t;
      total_error += GAdb2*t*t;
    } else if ( e <= DELTA )
      credit[k] = 0.0;
    else {
      t = e - DELTA;
      credit[k] = -GI * t;
      total_error += GIdb2*t*t;
    }
    
  }
  
  return( total_error );
}


void user_init()
{
  xor_set_random_init_range(0.5);

  xor_set_learning_rate(0.06); 
}

